#include <torch/extension.h>
#include <torch/torch.h>
#include <vector>
#include <c10/cuda/CUDAGuard.h>

torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    // Ensure we're operating on the correct device
    const at::cuda::CUDAGuard device_guard(x.device());
    
    // Pre-allocate contiguous tensors and ensure alignment
    h0 = h0.to(x.device()).contiguous();
    c0 = c0.to(x.device()).contiguous();
    x = x.contiguous();

    auto out = x;
    auto hn = h0.clone();
    auto cn = c0.clone();

    const size_t num_layers = lstm_weights_ih.size();
    
    // Pre-allocate and prepare all layer parameters in contiguous memory
    std::vector<torch::Tensor> weights_ih_contiguous;
    std::vector<torch::Tensor> weights_hh_contiguous;
    std::vector<torch::nn::LSTM> lstm_layers;
    
    weights_ih_contiguous.reserve(num_layers);
    weights_hh_contiguous.reserve(num_layers);
    lstm_layers.reserve(num_layers);

    // Prepare aligned and contiguous weights
    for (size_t i = 0; i < num_layers; ++i) {
        weights_ih_contiguous.push_back(lstm_weights_ih[i].contiguous());
        weights_hh_contiguous.push_back(lstm_weights_hh[i].contiguous());
        
        int64_t input_size = weights_ih_contiguous[i].size(1);
        int64_t hidden_size = weights_hh_contiguous[i].size(1);
        
        lstm_layers.emplace_back(
            torch::nn::LSTMOptions(input_size, hidden_size)
            .num_layers(1)
            .batch_first(true)
            .bidirectional(false)
        );
        
        auto& layer = lstm_layers[i];
        layer->to(x.device());
        
        // Assign pre-aligned weights and biases
        layer->named_parameters()["weight_ih_l0"].copy_(weights_ih_contiguous[i]);
        layer->named_parameters()["weight_hh_l0"].copy_(weights_hh_contiguous[i]);
        layer->named_parameters()["bias_ih_l0"].copy_(lstm_biases_ih[i]);
        layer->named_parameters()["bias_hh_l0"].copy_(lstm_biases_hh[i]);
        layer->train(is_training);
    }

    // Process layers with coalesced memory access
    for (size_t i = 0; i < num_layers; ++i) {
        auto h_slice = hn.narrow(0, i, 1).contiguous();
        auto c_slice = cn.narrow(0, i, 1).contiguous();
        
        std::tuple<torch::Tensor, torch::Tensor> state_tuple = std::make_tuple(h_slice, c_slice);
        
        // Ensure input is contiguous before LSTM operation
        out = out.contiguous();
        auto output_and_state = lstm_layers[i]->forward(out, state_tuple);
        
        out = std::get<0>(output_and_state);
        auto h_n = std::get<0>(std::get<1>(output_and_state));
        auto c_n = std::get<1>(std::get<1>(output_and_state));
        
        // Use contiguous tensors for state updates
        hn.narrow(0, i, 1).copy_(h_n.contiguous());
        cn.narrow(0, i, 1).copy_(c_n.contiguous());
    }

    return hn.contiguous();
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "LSTM forward (CUDA)");
}